from collections.abc import Callable, Iterable
from typing import Any

RowResult = dict[str, str]
SheetResult = Iterable[RowResult]
MultiSheetResult = Iterable[tuple[int, SheetResult]]
ValueMapper = Callable[[Any], Any]



# Generated by Django 4.2.16 on 2024-10-03 09:03

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="Household",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=100)),
                ("data", models.JSONField(blank=True, default=dict)),
            ],
        ),
        migrations.CreateModel(
            name="Individual",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("data", models.JSONField(blank=True, default=dict)),
                (
                    "household",
                    models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to="demo.household"),
                ),
            ],
        ),
    ]



# mypy: disable-error-code="no-untyped-def"
from pathlib import Path
from typing import Any, Iterable

import pytest
from demo.factories import FieldsetFactory, FlexFieldFactory
from hope_flex_fields.models import Fieldset

from hope_smart_import.readers import open_xls, open_xls_multi
from hope_smart_import.shortcuts import validate_single, validate_xls_multi


@pytest.fixture
def xls_simple() -> Iterable:
    return open_xls(str((Path(__file__).parent / "data" / "r1.xlsx").absolute()))


@pytest.fixture
def xls_multi() -> Iterable:
    return open_xls_multi(str((Path(__file__).parent / "data" / "r1.xlsx").absolute()), [0, 1])


@pytest.fixture
def simple_validator(db: Any) -> Fieldset:
    fs = FieldsetFactory(name="Simple Validator")
    FlexFieldFactory(name="name", fieldset=fs)
    FlexFieldFactory(name="last_name", fieldset=fs)
    return fs


def test_validate_simple(xls_simple: Iterable, simple_validator: Fieldset):
    g1 = list(xls_simple)
    errors = validate_single(g1, simple_validator, fail_if_alien=False)
    assert not errors

    errors = validate_single(g1, simple_validator, fail_if_alien=True)
    assert errors == {
        1: {"-": ["Alien values found {'gender'}"]},
        2: {"-": ["Alien values found {'gender'}"]},
    }


def test_validate_xls_multi(xls_multi: Iterable, simple_validator) -> None:
    errors = validate_xls_multi(xls_multi, [simple_validator, simple_validator], fail_if_alien=True)
    assert errors == {
        "1:Simple Validator": {
            1: {"-": ["Alien values found {'gender'}"]},
            2: {"-": ["Alien values found {'gender'}"]},
        },
        "2:Simple Validator": {
            1: {"-": ["Alien values found {'gender'}"]},
            2: {"-": ["Alien values found {'gender'}"]},
        },
    }



from collections.abc import Callable
from pathlib import Path
from typing import Any, TYPE_CHECKING

import pytest

from hope_smart_import.readers import open_csv, open_xls, open_xls_multi

if TYPE_CHECKING:
    from hope_smart_import.types import ValueMapper


@pytest.fixture
def xls() -> str:
    return str((Path(__file__).parent / "data" / "r1.xlsx").absolute())


@pytest.fixture
def csv() -> str:
    return str((Path(__file__).parent / "data" / "r1.csv").absolute())


@pytest.mark.parametrize("start_at", [0, 1])
def test_read_csv_headers(csv: str, start_at: int) -> None:
    for e in open_csv(csv, start_at=start_at, headers=True):
        assert list(e.keys()) == ["name", "last_name", "gender"]


@pytest.mark.parametrize("start_at", [0, 1])
def test_read_csv_no_headers(csv: str, start_at: int) -> None:
    for e in open_csv(csv, start_at=start_at, headers=False):
        assert list(e.keys()) == ["column1", "column2", "column3"]


@pytest.mark.parametrize(("mapper", "validator"), [(str.upper, str.isupper), (str.lower, str.islower)])
def test_read_csv_value_mapper(csv: str, mapper: "ValueMapper", validator: Callable[[Any], bool]) -> None:
    for row in open_csv(csv, headers=True, value_mapper=mapper):
        assert all(map(validator, row.values()))


@pytest.mark.parametrize("start_at", [0, 1])
def test_read_xls_headers(xls: str, start_at: int) -> None:
    for e in open_xls(xls, start_at=start_at):
        assert list(e.keys()) == ["name", "last_name", "gender"]


@pytest.mark.parametrize("start_at", [0, 1])
def test_read_xls_no_headers(xls: str, start_at: int) -> None:
    for e in open_xls(xls, start_at=start_at, headers=False):
        assert list(e.keys()) == ["column1", "column2", "column3"]


@pytest.mark.parametrize(("mapper", "validator"), [(str.upper, str.isupper), (str.lower, str.islower)])
def test_read_xls_value_mapper(xls: str, mapper: "ValueMapper", validator: Callable[[Any], bool]) -> None:
    for row in open_xls(xls, headers=True, value_mapper=mapper):
        assert all(map(validator, row.values()))


@pytest.mark.parametrize("start_at", [0, 1])
def test_read_multi_xls_headers(xls: str, start_at: int) -> None:
    ret = []
    for e, sh in open_xls_multi(xls, sheets=[0, 1], start_at=start_at, headers=True):
        ret.extend([(e, row) for row in sh])
    if start_at == 0:
        # Sheet 0 - Row 0
        assert ret[0] == (0, {"name": "John", "last_name": "Doe", "gender": "M"})
        # Sheet 0 - Row 1
        assert ret[1] == (0, {"name": "Jane", "last_name": "Doe", "gender": "F"})
        # Sheet 1 - Row 0
        assert ret[2] == (1, {"name": "John1", "last_name": "Doe1", "gender": "m"})
        # Sheet 1 - Row 1
        assert ret[3] == (1, {"name": "Jane1", "last_name": "Doe1", "gender": "f"})
    if start_at == 1:
        # Sheet 0 - Row 0
        assert ret[0] == (0, {"name": "Jane", "last_name": "Doe", "gender": "F"})
        # Sheet 1 - Row 1
        assert ret[1] == (1, {"name": "Jane1", "last_name": "Doe1", "gender": "f"})


@pytest.mark.parametrize(("mapper", "validator"), [(str.upper, str.isupper), (str.lower, str.islower)])
def test_read_multi_xls_value_mapper(xls: str, mapper: "ValueMapper", validator: Callable[[Any], bool]) -> None:
    for _, sheet in open_xls_multi(xls, sheets=[0, 1], headers=True, value_mapper=mapper):
        for row in sheet:
            assert all(map(validator, row.values()))



from django.db import models


class Household(models.Model):
    name = models.CharField(max_length=100)
    data = models.JSONField(default=dict, blank=True)

    def __str__(self):
        return self.name


class Individual(models.Model):
    household = models.ForeignKey(Household, on_delete=models.CASCADE)
    data = models.JSONField(default=dict, blank=True)

    def __str__(self):
        return f"{self.household.name} - {self.pk}"



# mypy: disable-error-code="no-untyped-def"
from pathlib import Path
from typing import TYPE_CHECKING, Any

import pytest
from demo.factories import FieldDefinitionFactory, FieldsetFactory, FlexFieldFactory
from django import forms
from hope_flex_fields.models import Fieldset

from hope_smart_import.readers import open_xls, open_xls_multi
from hope_smart_import.shortcuts import validate_single, validate_xls_multi

if TYPE_CHECKING:
    from hope_smart_import.types import MultiSheetResult, SheetResult


@pytest.fixture
def xls_rdi_simple() -> "SheetResult":
    return open_xls(str((Path(__file__).parent / "data" / "rdi1.xlsx").absolute()))


@pytest.fixture
def xls_rdi() -> "MultiSheetResult":
    return open_xls_multi(str((Path(__file__).parent / "data" / "rdi1.xlsx").absolute()), [0, 1])


@pytest.fixture
def xls_missing_master() -> "MultiSheetResult":
    return open_xls_multi(str((Path(__file__).parent / "data" / "missing_master.xlsx").absolute()), [0, 1])


@pytest.fixture
def hh_validator(db: Any) -> Fieldset:
    fs = FieldsetFactory(name="household")
    FlexFieldFactory(name="household_id", fieldset=fs)
    FlexFieldFactory(name="consent_h_c", fieldset=fs)
    FlexFieldFactory(name="country_origin_h_c", fieldset=fs)
    FlexFieldFactory(name="country_h_c", fieldset=fs)
    FlexFieldFactory(name="admin1_h_c", fieldset=fs)
    FlexFieldFactory(name="admin2_h_c", fieldset=fs)
    FlexFieldFactory(name="size_h_c", fieldset=fs)
    FlexFieldFactory(name="hh_latrine_h_f", fieldset=fs)
    FlexFieldFactory(name="hh_electricity_h_f", fieldset=fs)
    FlexFieldFactory(name="registration_method_h_c", fieldset=fs)
    FlexFieldFactory(name="collect_individual_data_h_c", fieldset=fs)
    FlexFieldFactory(name="name_enumerator_h_c", fieldset=fs)
    FlexFieldFactory(name="org_enumerator_h_c", fieldset=fs)
    FlexFieldFactory(name="consent_sharing_h_c", fieldset=fs)
    FlexFieldFactory(name="first_registration_date_h_c", fieldset=fs)
    return fs


@pytest.fixture
def simple_validator(db: Any) -> Fieldset:
    fs = FieldsetFactory()
    FlexFieldFactory(name="name", fieldset=fs)
    FlexFieldFactory(name="last_name", fieldset=fs)
    return fs


@pytest.fixture
def ind_validator(db: Any) -> Fieldset:
    fs = FieldsetFactory(name="individual")
    FlexFieldFactory(name="household_id", fieldset=fs)
    FlexFieldFactory(name="relationship_i_c", fieldset=fs)
    FlexFieldFactory(name="full_name_i_c", fieldset=fs)
    FlexFieldFactory(name="given_name_i_c", fieldset=fs)
    FlexFieldFactory(name="middle_name_i_c", fieldset=fs)
    FlexFieldFactory(name="family_name_i_c", fieldset=fs)
    FlexFieldFactory(name="photo_i_c", fieldset=fs)
    FlexFieldFactory(
        name="gender_i_c",
        fieldset=fs,
        definition=FieldDefinitionFactory(field_type=forms.ChoiceField),
        attrs={"choices": [["FEMALE", "FEMALE"], ["MALE", "MALE"]]},
    )

    FlexFieldFactory(
        name="birth_date_i_c",
        fieldset=fs,
        definition=FieldDefinitionFactory(field_type=forms.DateField),
    )
    FlexFieldFactory(name="estimated_birth_date_i_c", fieldset=fs)
    FlexFieldFactory(name="national_id_no_i_c", fieldset=fs)
    FlexFieldFactory(name="national_id_photo_i_c", fieldset=fs)
    FlexFieldFactory(name="national_id_issuer_i_c", fieldset=fs)
    FlexFieldFactory(name="phone_no_i_c", fieldset=fs)
    FlexFieldFactory(name="primary_collector_id", fieldset=fs)
    FlexFieldFactory(name="alternate_collector_id", fieldset=fs)
    FlexFieldFactory(name="first_registration_date_i_c", fieldset=fs)
    FlexFieldFactory(
        name="disability_i_c",
        fieldset=fs,
        definition=FieldDefinitionFactory(field_type=forms.ChoiceField),
        attrs={"choices": [["not disabled", "not disabled"], ["disabled", "disabled"]]},
    )
    return fs


def test_validate_simple(xls_rdi_simple: "SheetResult", hh_validator: Fieldset, ind_validator: Fieldset) -> None:
    hh_validator.set_primary_key_col("household_id")
    ind_validator.set_master(hh_validator, "household_id")

    errors = validate_single(xls_rdi_simple, hh_validator, fail_if_alien=True)

    assert errors == {}


def test_validate_multi(xls_rdi: "MultiSheetResult", hh_validator: Fieldset, ind_validator: Fieldset) -> None:
    errors = validate_xls_multi(xls_rdi, [hh_validator, ind_validator], fail_if_alien=True)
    assert errors == {"1:household": {}, "2:individual": {}}


def test_validate_master_detail(xls_rdi: "MultiSheetResult", hh_validator: Fieldset, ind_validator: Fieldset) -> None:
    hh_validator.set_primary_key_col("household_id")
    ind_validator.set_master(hh_validator, "household_id")

    errors = validate_xls_multi(xls_rdi, [hh_validator, ind_validator], fail_if_alien=True)
    assert errors == {"1:household": {}, "2:individual": {}}


def test_validate_missing_master(
    xls_missing_master: "MultiSheetResult",
    hh_validator: Fieldset,
    ind_validator: Fieldset,
) -> None:
    hh_validator.set_primary_key_col("household_id")
    ind_validator.set_master(hh_validator, "household_id")

    errors = validate_xls_multi(xls_missing_master, [hh_validator, ind_validator], fail_if_alien=True)

    assert errors["2:individual"] == {2: {"-": ["'missing' not found in master"]}}
    assert errors["1:household"] == {}



from typing import TYPE_CHECKING, Any
from collections.abc import Generator

from hope_flex_fields.models import DataChecker, Fieldset
from hope_flex_fields.models.base import ValidatorMixin

if TYPE_CHECKING:
    from .types import MultiSheetResult, SheetResult


def validate_single(
    g: "SheetResult",
    checker: ValidatorMixin,
    *,
    include_success: bool = False,
    fail_if_alien: bool = False,
) -> Generator[dict[str, Any]]:
    return checker.validate(g, include_success=include_success, fail_if_alien=fail_if_alien)


def validate_xls_multi(
    g: "MultiSheetResult",
    checkers: list[DataChecker | Fieldset],
    include_success: bool = False,
    fail_if_alien: bool = False,
) -> dict[str, list[dict[str, Any]]]:
    errors = {}
    for sheet_index, sheet_generator in g:
        checker = checkers[sheet_index]
        errors[f"{sheet_index + 1}:{checker.name}"] = checker.validate(
            sheet_generator,
            include_success=include_success,
            fail_if_alien=fail_if_alien,
        )
    return errors



from typing import Any, Optional

import factory
from django import forms
from django.contrib.auth.models import Group, Permission, User
from factory.django import DjangoModelFactory
from factory.faker import Faker
from hope_flex_fields.models import DataChecker, Fieldset, FlexField
from hope_flex_fields.models.definition import FieldDefinition
from strategy_field.utils import fqn

from hope_smart_import.models import Configuration


class FieldsetFactory(DjangoModelFactory):
    extends = None

    class Meta:
        model = Fieldset


class FieldDefinitionFactory(DjangoModelFactory):
    field_type = fqn(forms.CharField)

    class Meta:
        model = FieldDefinition
        django_get_or_create = ("field_type",)


class FlexFieldFactory(DjangoModelFactory):
    name = factory.Sequence(lambda n: f"FlexField {n}")
    fieldset = factory.SubFactory(Fieldset)
    definition = factory.SubFactory(FieldDefinitionFactory)

    class Meta:
        model = FlexField


class DataCheckerFactory(DjangoModelFactory):
    name = factory.Sequence(lambda n: f"DataChecker {n}")

    class Meta:
        model = DataChecker

    @factory.post_generation  # type: ignore[misc]
    def fieldsets(self, create: bool, extracted: list[str], **kwargs: Any) -> None:
        if not create:
            # Simple build, do nothing.
            return

        if extracted:
            for perm_name in extracted:
                app, perm = perm_name.split(".")
                self.fieldsets.add(Permission.objects.get(content_type__app_label=app, codename=perm))


class ConfigurationFactory(DjangoModelFactory):
    name = factory.Sequence(lambda n: f"Configuration {n}")

    class Meta:
        model = Configuration


class UserFactory(DjangoModelFactory):
    username = factory.Sequence(lambda n: f"User {n}")
    email = factory.Sequence(lambda n: f"email{n}@example.com")

    class Meta:
        model = User


class GroupFactory(DjangoModelFactory):
    name = Faker("name")

    class Meta:
        model = Group
        django_get_or_create = ("name",)

    @factory.post_generation  # type: ignore[misc]
    def permissions(self, create: bool, extracted: list[str], **kwargs: Any) -> None:
        if not create:
            # Simple build, do nothing.
            return

        if extracted:
            for perm_name in extracted:
                app, perm = perm_name.split(".")
                self.permissions.add(Permission.objects.get(content_type__app_label=app, codename=perm))


class user_grant_permission:  # noqa
    def __init__(self, user: User, permissions: Optional[list[str]] = None):
        self.user = user
        self.permissions = permissions
        self.group = None

    def __enter__(self):
        if hasattr(self.user, "_group_perm_cache"):
            del self.user._group_perm_cache
        if hasattr(self.user, "_perm_cache"):
            del self.user._perm_cache
        self.group = GroupFactory(permissions=self.permissions or [])
        self.user.groups.add(self.group)

    def __exit__(self, *exc_info):
        if self.group:
            self.user.groups.remove(self.group)
            self.group.delete()

    def start(self):
        """Activate a patch, returning any created mock."""
        return self.__enter__()

    def stop(self):
        """Stop an active patch."""
        return self.__exit__()



import csv
from collections.abc import Callable
from itertools import islice
from typing import TYPE_CHECKING, Any, Iterable

import openpyxl

if TYPE_CHECKING:
    from openpyxl.workbook.workbook import Workbook
    from openpyxl.worksheet.worksheet import Worksheet

    from .types import MultiSheetResult, SheetResult, ValueMapper


def identity(x: Any) -> Any:
    return x


def _read_worksheet(
    sheet: "Worksheet",
    start_at: int,
    headers: bool,
    value_mapper: Callable[[Any], Any] = identity,
) -> Iterable[dict[str, Any]]:
    rows = sheet.rows

    field_names = None
    generate_field_names = False
    if headers:
        field_names = [str(cell.value) for cell in next(rows)]
    else:
        generate_field_names = True

    rows = islice(rows, start_at, None)
    for row in rows:
        if generate_field_names:
            field_names = [f"column{d + 1}" for d in range(len(row))]

        yield dict(zip(field_names, (value_mapper(cell.value) for cell in row), strict=True))


def open_xls(
    filepath: str,
    *,
    sheet_index: int = 0,
    start_at: int = 0,
    headers: bool = True,
    value_mapper: "ValueMapper" = identity,
) -> "SheetResult":
    wb: "Workbook" = openpyxl.load_workbook(filepath)
    sh: "Worksheet" = wb.worksheets[sheet_index]
    yield from _read_worksheet(sh, start_at=start_at, headers=headers, value_mapper=value_mapper)


def open_xls_multi(
    filepath: str,
    sheets: list[int] = (0,),
    start_at: list[int] | int = 0,
    headers: list[bool] | bool = True,
    value_mapper: "ValueMapper" = identity,
) -> "MultiSheetResult":
    wb: "Workbook" = openpyxl.load_workbook(filepath)
    for si in sheets:
        sh: "Worksheet" = wb.worksheets[si]
        start_at = start_at if isinstance(start_at, int) else start_at[si]
        headers = headers if isinstance(headers, bool) else headers[si]
        yield (
            si,
            _read_worksheet(sh, start_at=start_at, headers=headers, value_mapper=value_mapper),
        )


def open_csv(
    filepath: str,
    *,
    start_at: int = 0,
    headers: bool = False,
    value_mapper: "ValueMapper" = identity,
) -> "SheetResult":
    with open(filepath) as f:
        if headers:
            reader = csv.DictReader(f)
            reader = islice(reader, start_at, None)
            for row in reader:
                yield {k: value_mapper(v) for k, v in row.items()}
        else:
            reader = csv.reader(f)
            reader = islice(reader, start_at, None)
            for row in reader:
                field_names = [f"column{d + 1}" for d in range(len(row))]
                yield dict(zip(field_names, map(value_mapper, row), strict=True))
